/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.prognose;

import java.text.DateFormat;
import java.util.Collections;
import java.util.Date;

import com.bitctrl.Constants;
import com.bitctrl.util.Interval;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.iav.gllib.gllib.dav.GanglinieMQ;
import de.bsvrz.iav.gllib.gllib.dav.Ganglinienprognose;
import de.bsvrz.iav.gllib.gllib.dav.GlProgAnfrage;
import de.bsvrz.iav.gllib.gllib.dav.GlProgAntwortEvent;
import de.bsvrz.iav.gllib.gllib.dav.GlProgAntwortListener;
import de.bsvrz.ibv.simbaustelle.simbaustelle.BaustellenSimulation;
import de.bsvrz.ibv.simbaustelle.simbaustelle.BssMessage;
import de.bsvrz.ibv.simbaustelle.verwaltung.BaustellenSchnittstelle;
import de.bsvrz.sys.funclib.bitctrl.daf.BetriebsmeldungDaten;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensendeException;
import de.bsvrz.sys.funclib.bitctrl.modell.kalender.objekte.EreignisTyp;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.bitctrl.util.UndefiniertException;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * eine Klasse zur Ermittlung des zufließenden Verkehrs zu einer Baustelle.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class ZufliessenderVerkehr implements ZufliessenderVerkehrSchnittstelle, GlProgAntwortListener {

	/**
	 * der Logger für Debugausgaben.
	 */
	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * globaler statischer Zähler zur Identifikation von Anfragen für den
	 * zufließenden Verkehr.
	 */
	private static int nextAnfrageId = 1;

	/**
	 * liefert die Markierung für den nächsten Anfrageauftrag an die
	 * Ganglinienprognose.
	 *
	 * @return das Zeichen
	 */
	private static String getNextAuftragsZeichen() {
		return ZufliessenderVerkehr.class.getName() + "." + (ZufliessenderVerkehr.nextAnfrageId++);
	}

	/**
	 * die Ganglinie des Messquerschnitts unmittelbar vor der Baustelle.
	 */
	private GanglinieMQ gangLinie;

	/**
	 * der vor der Baustelle liegende Messquerschnitt.
	 */
	private final MessQuerschnittAllgemein mq;

	/**
	 * die Markierung für Anfragen an die Ganglinienprognose.
	 */
	private final String auftragsZeichen;

	/**
	 * erzeugt eine Klasse zur Bereitstellung des zur Baustelle zufließenden
	 * Verkehrs.Der zufließende Verkehr ergibt sich aus der Ganglinie des
	 * unmittelbar vor der Baustelle liegenden Messquerschnitts.
	 *
	 * @param verbindung
	 *            die verwendete Datenverteilerverbindung
	 * @param baustelle
	 *            die Baustelle für die der zufließende Verkehr ermittelt werden
	 *            soll.
	 * @param berechnungsMq
	 *            optional vorgegebener Mq vor der Baustelle.
	 * @param zeitRaum
	 *            der Zeitraum, für die eine Ermittlung erfolgen soll und damit
	 *            eine Ganglinie bestimmt werden muß
	 * @throws UndefiniertException
	 *             es konnte kein geeigneter Messquerschnitt oder keine
	 *             Ganglinie ermittelt werden.
	 */
	public ZufliessenderVerkehr(final ClientDavInterface verbindung, final BaustellenSchnittstelle baustelle,
			final MessQuerschnittAllgemein berechnungsMq, final BerechnungsZeitRaumSchnittstelle zeitRaum)
					throws UndefiniertException {

		if (berechnungsMq != null) {
			mq = berechnungsMq;
		} else {
			mq = baustelle.getMessquerschnittDavor();
		}
		final Ganglinienprognose prognose = Ganglinienprognose.getInstanz();
		prognose.addAntwortListener(this);

		final GlProgAnfrage anfrage = new GlProgAnfrage(mq,
				new Interval(zeitRaum.getStartZeitpunkt(), zeitRaum.getEndeZeitpunkt()), true);

		final SystemObject baustellenEreignisTyp = verbindung.getDataModel()
				.getObject(BaustellenSimulation.getBstErgTypPid());
		if (baustellenEreignisTyp != null) {
			anfrage.getEreignisTypen().add(new EreignisTyp(baustellenEreignisTyp));
		}

		auftragsZeichen = ZufliessenderVerkehr.getNextAuftragsZeichen();

		try {
			synchronized (this) {
				ZufliessenderVerkehr.LOGGER.fine("Anfrage an Ganglinienprognose versenden: " + auftragsZeichen);
				prognose.sendeAnfrage(auftragsZeichen, Collections.singletonList(anfrage));
				this.wait(Constants.MILLIS_PER_MINUTE * 2);
			}
		} catch (final InterruptedException e) {
			LogTools.log(ZufliessenderVerkehr.LOGGER, new BetriebsmeldungDaten(baustelle.getObjekt()),
					BssMessage.WarningGanglinienAbfrage, baustelle, zeitRaum, e.getMessage());
		} catch (final DatensendeException e) {
			LogTools.log(ZufliessenderVerkehr.LOGGER, new BetriebsmeldungDaten(baustelle.getObjekt()),
					BssMessage.WarningGanglinienAbfrage, baustelle, zeitRaum, e.getMessage());
		}

		prognose.removeAntwortListener(this);
	}

	@Override
	public final void antwortEingetroffen(final GlProgAntwortEvent e) {
		ZufliessenderVerkehr.LOGGER.fine("Antwort von Ganglinienprognose eingetroffen: " + e.getAbsenderZeichen());
		if (e.getAbsenderZeichen().equals(auftragsZeichen)) {
			ZufliessenderVerkehr.LOGGER.fine("Verarbeite Antwort von Ganglinienprognose");
			for (final GanglinieMQ gl : e.getGanglinien()) {
				final MessQuerschnittAllgemein glMq = gl.getMessQuerschnitt();
				if (glMq == null) {
					continue;
				}
				if (glMq.equals(mq)) {
					gangLinie = gl;
					if (BaustellenSimulation.isPolyLinie()) {
						gangLinie.setApproximationDaK(GanglinieMQ.APPROX_POLYLINE);
					}
				}
			}

			synchronized (this) {
				this.notifyAll();
			}
		}
	}

	@Override
	public final long getFzg(final long zs) throws UndefiniertException {

		if (gangLinie == null) {
			throw new UndefiniertException(
					"Zufliessender Verkehr: Es konnte keine Ganglinie für MQ:" + mq + " ermittelt werden");
		}
		final Double wert = gangLinie.getStuetzstelle(zs).getWert().getQB();
		if (wert == null) {
			throw new UndefiniertException("In der Ganglinie des MQ: " + mq + " existiert kein Wert für den Zeitpunkt: "
					+ DateFormat.getDateTimeInstance().format(new Date(zs)));
		}
		return wert.longValue();
	}
}
