/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.simbaustelle;

import java.util.List;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.ibv.simbaustelle.parameter.BssParameterObserver;
import de.bsvrz.ibv.simbaustelle.parameter.IBssParameterObserver;
import de.bsvrz.ibv.simbaustelle.verwaltung.BaustellenVerwaltung;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.application.StandardApplicationRunner;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.SystemObjekt;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.Baustelle;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.VerkehrModellNetz;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.parameter.PdSituationsEigenschaften;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.parameter.PdSituationsEigenschaften.Daten;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList.Argument;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;

/**
 * Die Hauptklasse der Baustellensimulation.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public final class BaustellenSimulation implements StandardApplication {

	/**
	 * das Kommandozeilenargument zur Definition des Netzes f&uuml;r das die
	 * Baustellensimulation ausgeführt werden soll. Das Argument ist eine PID,
	 * die ein Objekt vom Typ "VerkehrsModellNetz" beschreibt.
	 */
	private static final String ARGUMENT_NAME_NETZ = "-netz=";

	/** das Kommandozeilenargument für den Testbetrieb. */
	private static final String ARGUMENT_NAME_CHECK = "-check";

	/** das Kommandozeilenargument für erlaubte Lücken in einer Ganglinie. */
	private static final String ARGUMENT_NAME_ALLOW_GAPS = "-allowGaps";

	/**
	 * das Kommandozeilenargument für die Zufluss-MQ-Suche im Afo-strengen
	 * Modus.
	 */
	private static final String ARGUMENT_NAME_MQ_AFO_STRICT = "-mqAfoStrict";

	/**
	 * das Kommandozeilenargument zur Definition der PID für die Ermittlung des
	 * Baustellen-Ereignistyps.
	 */
	private static final String ARGUMENT_NAME_BST_ERG_TYPE = "-bstErgTyp=ergTyp.baustelle";

	/**
	 * das Kommandozeilenargument zur Festlegung der Fahrstreifenanzehl, wenn
	 * keine definiert ist.
	 */
	private static final String ARGUMENT_NAME_DEFAULT_FAHRSTREIFEN_ANZAHL = "-fahrstreifen=2";

	/**
	 * das Kommandozeilenargument zur Festlegung, ob Ganglinien als Polylinie
	 * ausgewertet werden sollen.
	 */
	private static final String ARGUMENT_NAME_GANGLINIE_POLYLINIE = "-polyLinie";

	/**
	 * das Kommandozeilenargument zur Festlegung, ob die Zeitbereichsbildung
	 * streng nach Afo erfolgen soll.
	 */
	private static final String ARGUMENT_NAME_ZEITBEREICH_STRICT = "-zeitBereichStrict";

	/**
	 * das Kommandozeilenargument beschränkt den maximalen Zeitraum für manuelle
	 * Simulationen auf die Anzahl der festgelegten Tage. Der Standardwert ist
	 * 0, was als "keine Beschränkung" interpretiert wird.
	 */
	private static final String ARGUMENT_MAX_HORIZONT_MANUELL = "-maxHorizontManuell";

	/**
	 * der Logger für Debugausgaben.
	 */
	private static Debug logger;

	/**
	 * die PID des VerkehrsmodellNetz für das eine Simulation erfolgen soll.
	 */
	private static String bstErgTypPid = "ergTyp.baustelle";

	/**
	 * die Standardanzahl für Fahrstreifen, wenn keiner ermittelt werden konnte.
	 */
	private static int defaultFahrstreifenAnzahl = 2;

	/**
	 * die Standardanzahl für Fahrstreifen, wenn keiner ermittelt werden konnte.
	 */
	private static boolean polyLinie;

	/**
	 * liefert die PID zur Bestimmung des Baustellenereignistyps.
	 *
	 * @return die PID
	 */
	public static String getBstErgTypPid() {
		return BaustellenSimulation.bstErgTypPid;
	}

	/**
	 * liefert die Standard-Fahrstreofenanzahl.
	 *
	 * @return die Anzahl
	 */
	public static int getDefaultFahrstreifenAnzahl() {
		return BaustellenSimulation.defaultFahrstreifenAnzahl;
	}

	/**
	 * liefert die globale Instanz der Baustellensimulation.
	 *
	 * @return die Instanz
	 */
	public static BaustellenSimulation getInstance() {
		return BaustellenSimulation.INSTANCE;
	}

	/**
	 * liefert Auskunft, ob die per Ganglinie erhaltenen Ergebnisse zwangsweise
	 * als Polylinie verwendet werden.
	 *
	 * @return <code>true</code>, wenn mittels PolyLinie berechnet werden soll
	 */
	public static boolean isPolyLinie() {
		return BaustellenSimulation.polyLinie;
	}

	/**
	 * die Hauptfunktion der Softwareeinheit.
	 *
	 * @param args
	 *            die Kommadozeilenparameter.
	 */
	public static void main(final String[] args) {
		ObjektFactory.getInstanz().registerStandardFactories();
		StandardApplicationRunner.run(BaustellenSimulation.getInstance(), args);
	}

	/**
	 * die PID des VerkehrsmodellNetz für das eine Simulation erfolgen soll.
	 */
	private String netzPid;

	/**
	 * die Parameterverwaltung zur Bereitstellung der grundlegenden Parameter
	 * der Baustellensimulation des Netzes.
	 */
	private IBssParameterObserver parameter;

	/**
	 * die verwendete Datenverteilerverbindung.
	 */
	private ClientDavInterface verbindung;

	/** Baustellensimulation nur im Testmodus ausführen. */
	private boolean checkMode;

	/** Lücken in einer Ganglinie sind erlaubt. */
	private boolean allowGaps;

	/**
	 * MQ für die Zuflussberechnung streng nach Anwenderanforderungen suchen.
	 * (Standard ist die ISIS-Methode).
	 */
	private boolean mqAfoStrict;

	/** Bildung des Zeitbereichs für die Simulation streng nach Afo? */
	private boolean zeitbereichAfoStrict;

	/** der maximale Zeitbereich für manuelle Simulationen in Millisekunden. */
	private long maxHorizontManuellMsecs;

	/** globale Instanz der Applikation. */
	private static final BaustellenSimulation INSTANCE = new BaustellenSimulation();

	/** privater Konstruktor. */
	private BaustellenSimulation() {
		// Konstruktor wird nur intern aufgerufen
	}

	/**
	 * beendet das Programm und zeigt die zu verwendenden Parameter an.
	 */
	private void exitWithUsage() {
		System.out.println();
		System.out.println("Aufrufparameter:");
		System.out.println();
		System.out.println(
				"   -netz=<pid>  die Pid des VerkehrsModellNetz, für das die Baustellensimulation erfolgen soll");

		System.exit(-1);
	}

	/**
	 * liefert das Untermodul zr Bereitstellung der Parameter für die
	 * Simulation.
	 *
	 * @return des Modul
	 */
	public IBssParameterObserver getParameter() {
		return parameter;
	}

	/**
	 * liefert die verwendete Datenverteilerverbindung.
	 *
	 * @return die Verbindung
	 */
	public ClientDavInterface getVerbindung() {
		return verbindung;
	}

	/**
	 * Die Funktion startet die Untermdule zur Verwaltung der
	 * Simulationsparameter und das Verwaltungsmodul.
	 *
	 * @see de.bsvrz.sys.funclib.application.StandardApplication#initialize(de.bsvrz.dav.daf.main.ClientDavInterface)
	 */
	@Override
	public void initialize(final ClientDavInterface connection) throws Exception {

		MessageSender.getInstance()
		.setApplicationLabel(getClass().getSimpleName() + "_" + connection.getLocalApplicationObject().getId());

		BaustellenSimulation.logger.finer("Starte Verwaltungsmodul");

		this.verbindung = connection;
		ObjektFactory.getInstanz().setVerbindung(verbindung);

		/* TEST */

		if (checkMode) {
			final List<SystemObjekt> bsts = ObjektFactory.getInstanz().bestimmeModellobjekte("typ.baustelle");

			System.err.println("Ermittle MQ vor den vor vorhandenen Baustellen");
			System.err.println("==============================================\n");

			for (final SystemObjekt bst : bsts) {
				MessQuerschnittAllgemein messquerschnittDavor = null;
				final PdSituationsEigenschaften situation = ((Baustelle) bst).getSituationsEigenschaften();
				if (situation != null) {
					System.err.println("Teste: " + bst);
					final Daten datum = situation.abrufenDatum();
					if (datum != null) {
						messquerschnittDavor = datum.getMessquerschnittDavor();
					}
				}
				System.err.println(bst.getPid() + "--> " + messquerschnittDavor);
			}

			System.err.println("\n\nDas war nur ein Test\n\n");
			System.exit(0);
		}

		/* TEST-ENDE */

		final ConfigurationObject netzObj = (ConfigurationObject) connection.getDataModel().getObject(netzPid);
		if (netzObj == null) {
			throw new IllegalArgumentException("Die PID: \"" + netzPid + "\" beschreibt kein VerkehrsModellNetz");
		}

		parameter = new BssParameterObserver(connection, netzObj);
		final VerkehrModellNetz netz = (VerkehrModellNetz) ObjektFactory.getInstanz().getModellobjekt(netzObj);
		new BaustellenVerwaltung(this, netz).start();
	}

	/**
	 * ermittelt, ob Lücken in einer Ganglinie zulässig sind.
	 *
	 * @return den Status
	 */
	public boolean isAllowGaps() {
		return allowGaps;
	}

	/**
	 * ermittelt, ob der MQ für den Zufluss einer Baustelle streng nach
	 * Anwenderforderungen oder nach ISIS-Methode gesucht werden soll.
	 *
	 * @return der Zustand
	 */
	public boolean isMqAfoStrict() {
		return mqAfoStrict;
	}

	@Override
	public void parseArguments(final ArgumentList argumentList) throws Exception {
		BaustellenSimulation.logger = Debug.getLogger();
		try {
			netzPid = argumentList.fetchArgument(BaustellenSimulation.ARGUMENT_NAME_NETZ).asNonEmptyString();
		} catch (final IllegalArgumentException e) {
			System.out.println();
			System.out.println(e.getMessage());
			System.out.println("Programm wird beendet");
			exitWithUsage();
		}

		BaustellenSimulation.bstErgTypPid = argumentList.fetchArgument(BaustellenSimulation.ARGUMENT_NAME_BST_ERG_TYPE)
				.asNonEmptyString();

		BaustellenSimulation.defaultFahrstreifenAnzahl = argumentList
				.fetchArgument(BaustellenSimulation.ARGUMENT_NAME_DEFAULT_FAHRSTREIFEN_ANZAHL).intValue();

		BaustellenSimulation.polyLinie = argumentList
				.fetchArgument(BaustellenSimulation.ARGUMENT_NAME_GANGLINIE_POLYLINIE).booleanValue();

		if (argumentList.hasArgument(BaustellenSimulation.ARGUMENT_NAME_CHECK)) {
			checkMode = argumentList.fetchArgument(BaustellenSimulation.ARGUMENT_NAME_CHECK + "=").booleanValue();
		}

		if (argumentList.hasArgument(BaustellenSimulation.ARGUMENT_NAME_ALLOW_GAPS)) {
			allowGaps = argumentList.fetchArgument(BaustellenSimulation.ARGUMENT_NAME_ALLOW_GAPS + "=").booleanValue();
		}

		if (argumentList.hasArgument(BaustellenSimulation.ARGUMENT_NAME_MQ_AFO_STRICT)) {
			mqAfoStrict = argumentList.fetchArgument(BaustellenSimulation.ARGUMENT_NAME_MQ_AFO_STRICT + "=")
					.booleanValue();
		}

		if (argumentList.hasArgument(BaustellenSimulation.ARGUMENT_NAME_ZEITBEREICH_STRICT)) {
			zeitbereichAfoStrict = argumentList
					.fetchArgument(BaustellenSimulation.ARGUMENT_NAME_ZEITBEREICH_STRICT + "=").booleanValue();
		}

		if (argumentList.hasArgument(BaustellenSimulation.ARGUMENT_MAX_HORIZONT_MANUELL)) {
			maxHorizontManuellMsecs = argumentList
					.fetchArgument(BaustellenSimulation.ARGUMENT_MAX_HORIZONT_MANUELL + "=").longValue() * 24L * 3600L
					* 1000L;
		}

		if (argumentList.getUnusedArgumentCount() > 0) {
			System.out.println("Unerwartete Argumente:");
			Argument argument;
			while ((argument = argumentList.fetchNextArgument()) != null) {
				System.out.println(argument);
			}
			System.out.println("Programm wird beendet");
			exitWithUsage();
		}
	}

	/**
	 * bestimmt, ob die strikte Zeitbereichsdefinition der Afo verwendet wird.
	 *
	 * @return den Zustand
	 */
	public boolean isZeitbereichAfoStrict() {
		return zeitbereichAfoStrict;
	}

	/**
	 * liefert den maximalen Zeitbereich für manuelle Simulationen in
	 * Millisekunden. Ein Wert &lt;= 0 bedeutet, dass keine Beschränkung wirksam
	 * ist.
	 *
	 * @return den Wert
	 */
	public long getMaxHorizontManuellMsecs() {
		return maxHorizontManuellMsecs;
	}

}
