/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.verwaltung;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.bitctrl.Constants;

import de.bsvrz.ibv.simbaustelle.parameter.IBssParameterObserver;
import de.bsvrz.ibv.simbaustelle.prognose.FundamentalDiagramm;
import de.bsvrz.ibv.simbaustelle.simbaustelle.BaustellenSimulation;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensatzUpdateEvent;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensatzUpdateListener;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.StrassenSegment;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.parameter.PdBaustellenEigenschaften;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.parameter.PdSituationsEigenschaften;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.zustaende.BaustellenStatus;
import de.bsvrz.sys.funclib.bitctrl.util.SearchCycleException;
import de.bsvrz.sys.funclib.bitctrl.util.UndefiniertException;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Die Repräsentation der parametrierbaren Daten einer Baustelle.<br>
 * Die Klasse verwaltet die Einträge aus den Parameterdatensätze
 * "atg.situationsEigenschaften" und "atg.baustellenEigenschaften".
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class BaustellenDaten implements DatensatzUpdateListener, BaustellenDatenSchnittstelle {

	/** Hintergrund-Thread zur Verarbeitung neuer Messquerschnittdaten. */
	private static class MqUpdater extends Thread {

		/** die Auftragsliste. */
		private final Set<BaustellenDaten> reqList = new HashSet<>();
		/** Mutex zur Synchronisation der Auftragsliste. */
		private final Object lock = new Object();

		/** Standardkonstruktor. */
		MqUpdater() {
			super("Simbaustelle-MQ-Updater");
			setDaemon(true);
		}

		/**
		 * fügt einen neuen Auftrag hinzu.
		 *
		 * @param daten
		 *            die Daten des neuen Auftrags
		 */
		void addRequest(final BaustellenDaten daten) {
			synchronized (reqList) {
				reqList.add(daten);
				synchronized (lock) {
					lock.notifyAll();
				}
			}

			if (!isAlive()) {
				start();
			}
		}

		@Override
		public void run() {
			while (true) {
				BaustellenDaten daten = null;
				int remaining = 0;
				synchronized (reqList) {
					if (reqList.size() > 0) {
						daten = reqList.toArray(new BaustellenDaten[reqList.size()])[0];
						reqList.remove(daten);
						remaining = reqList.size();
						Debug.getLogger().finer(
								"Es liegen noch " + remaining + " Aktualisierungsaufträge für MQ von Baustellen vor");
					}
				}
				if (daten != null) {
					daten.aktualisiereMQDaten();
				}

				if (remaining <= 0) {
					synchronized (lock) {
						try {
							lock.wait(60000);
						} catch (final InterruptedException e) {
							Debug.getLogger().fine(e.getLocalizedMessage());
							// Schleife wird wiederholt und eventuelle Aufträge
							// vorfristig ausgeführt
						}
					}
				}
			}
		}
	}

	/** Thread zur Aktualisierung der MQ-Daten. */
	private static MqUpdater updater = new MqUpdater();

	/**
	 * die zugeordnete Baustelle.
	 */
	private final Baustelle baustelle;

	/**
	 * das sich aus der Liste der Straßensegmente ergebende aktuelle
	 * Fundamentaldiagramm des letzen Messquerschnitts der Naustelle.
	 */
	private final FundamentalDiagramm fundamentalDiagram;

	/**
	 * die Parameter der Baustellensimulation.
	 */
	private final IBssParameterObserver parameter;

	/**
	 * der sich aus der aktuellen Liste der Straßensegmente ergebende
	 * Meßquerschnitt vor der Baustelle. Dieser wird zur Berechnung des
	 * zufließenden Verkehrs genutzt.
	 */
	private MessQuerschnittAllgemein mqDavor;

	/**
	 * erzeugt eine Instanz der Klasse zur Speicherung der Baustellenparameter.
	 *
	 * @param baustelle
	 *            die zugeordnete Baustelle
	 * @param parameter
	 *            die Parameter der Baustellensimulation
	 */
	public BaustellenDaten(final Baustelle baustelle, final IBssParameterObserver parameter) {

		this.baustelle = baustelle;
		this.parameter = parameter;

		fundamentalDiagram = new FundamentalDiagramm();

		baustelle.getParameterDatensatz(PdSituationsEigenschaften.class).addUpdateListener(this);
		baustelle.getParameterDatensatz(PdBaustellenEigenschaften.class).addUpdateListener(this);

		BaustellenDaten.updater.addRequest(this);
	}

	/**
	 * aktualiesiert die für die Simulation verwendeten Messquerschnitte und das
	 * damit benötigte Fundamentaldiagramm. Die Aktualisierung erfolgt jedesmal,
	 * wenn sich der Parameterdatensatz der Situation ändert.
	 */
	private void aktualisiereMQDaten() {
		synchronized (fundamentalDiagram) {
			final Debug logger = Debug.getLogger();
			logger.finer("Aktualisiere MQ-Daten für Baustelle: " + baustelle.getObjekt());
			final List<StrassenSegment> segmente;
			try {
				logger.finest("Bestimme Segmente der Baustelle");
				segmente = getSegmente();
				// letzten MQ bestimmen und Fundamentaldiagramm ermitteln

				if (segmente.size() <= 0) {
					throw new UndefiniertException("Baustelle: " + baustelle + " hat keine zugeordneten Segmente");
				}

				final StrassenSegment segment = segmente.get(segmente.size() - 1);
				logger.finest("Bestimme Messquerschnitte der Baustelle");
				final List<MessQuerschnittAllgemein> mqs = segment.getMessquerschnitte();
				if (mqs.size() > 0) {
					final MessQuerschnittAllgemein mq = mqs.get(mqs.size() - 1);
					logger.finest("Setze Fundamentaldiagramn auf MQ: " + mq.getSystemObject());
					fundamentalDiagram.setMq(mq);
				}

				logger.finest("Ermittle MQ vor der Baustelle");
				if (BaustellenSimulation.getInstance().isMqAfoStrict()) {
					mqDavor = getSituationsDaten().getMessquerschnittDavor();
				} else {
					if (!segmente.isEmpty()) {
						final StrassenSegment firstSegment = segmente.get(0);
						final long startOffset = getSituationsDaten().getStartOffset();
						MessQuerschnittAllgemein mqVorBst = null;
						for (final MessQuerschnittAllgemein mq : firstSegment.getMessquerschnitte()) {
							final float mqOffset = mq.getOffset();
							if (mqOffset < startOffset) {
								if ((mqVorBst == null) || (mqVorBst.getOffset() < mqOffset)) {
									mqVorBst = mq;
								}
							} else {
								if ((mqDavor == null) || (mqDavor.getOffset() > mqOffset)) {
									mqDavor = mq;
								}
							}
						}
						if (mqVorBst != null) {
							mqDavor = mqVorBst;
						}
					}
				}

			} catch (@SuppressWarnings("unused") final UndefiniertException e) {
				logger.finest("Baustelle hat keine zugeordneten Segmente");
				fundamentalDiagram.setMq(null);
				mqDavor = null;
			} catch (final SearchCycleException e) {
				logger.warning(baustelle.getPid() + ":" + e.getLocalizedMessage(),
						e.getVisitedSegments().toArray(new StrassenSegment[0]));
			}
			logger.finer("MQ-Daten für Baustelle: " + baustelle.getObjekt() + " aktualisiert");
		}
	}

	@Override
	public void datensatzAktualisiert(final DatensatzUpdateEvent event) {
		if (event.getDatensatz() instanceof PdSituationsEigenschaften) {
			BaustellenDaten.updater.addRequest(this);
		}
	}

	@Override
	public final long getAnfangsOffset() throws UndefiniertException {
		return getSituationsDaten().getStartOffset();
	}

	/**
	 * liefert den aktuellen Datensatz mit den Baustellenparametern. Wenn keine
	 * gültigen Parameter verfügbar sind wird eine {@link UndefiniertException}
	 * geworfen.
	 *
	 * @return die Daten
	 * @throws UndefiniertException
	 *             es konnten keine gültigen Parameter ermittelt werden
	 */
	private PdBaustellenEigenschaften.Daten getBaustellenDaten() throws UndefiniertException {
		final PdBaustellenEigenschaften.Daten daten = baustelle.getParameterDatensatz(PdBaustellenEigenschaften.class)
				.getDatum();
		if ((daten == null) || (!daten.isValid())) {
			throw new UndefiniertException("Die Daten der Baustelle " + baustelle + " sind nicht bekannt");
		}
		return daten;
	}

	@Override
	public final long getEnde() throws UndefiniertException {
		return getSituationsDaten().getEndZeit();
	}

	@Override
	public final long getEndOffset() throws UndefiniertException {
		return getSituationsDaten().getEndOffset();
	}

	/**
	 * liefert den innerhalb der Baustelle aktuell ermittelten Messquerschnitt,
	 * der unmittelbar vor der Baustelle liegt.
	 *
	 * @return den Messquerschnitt
	 * @throws UndefiniertException
	 *             es konnte kein entsprechender Messquerschnitt ermittelt
	 *             werden
	 */
	public final MessQuerschnittAllgemein getMessquerschnittDavor() throws UndefiniertException {
		if (mqDavor == null) {
			throw new UndefiniertException(
					"Es konnte kein vor der Baustelle liegender Messquerschnitt ermittelt werden");
		}
		return mqDavor;
	}

	@Override
	public final long getRestKapazitaet(final long zs) throws UndefiniertException {
		if (fundamentalDiagram == null) {
			throw new UndefiniertException("Für den letzten Messquerschniit der Baustelle " + baustelle.getName()
			+ " ist kein Fundamentaldiagramm definiert");
		}
		long rest = 0L;
		if ((zs < getStart()) || (zs > getEnde())) {
			rest = (long) (fundamentalDiagram.getQ0() * (parameter.getFaktorQ0() / Constants.PERCENT));
		} else {
			rest = getBaustellenDaten().getRestKapazitaet();
		}
		return rest;
	}

	/**
	 * liefert die Liste der Strassensegmente auf denen die Baustelle liegt.
	 *
	 * @return die Liste der Segment
	 * @throws UndefiniertException
	 *             es konnten keine Segmente ermittelt werden, weil der
	 *             Parameterdatensatz nicht vorlag oder der Parameterdatensatz
	 *             keine Segmente enthält
	 */
	public final List<StrassenSegment> getSegmente() throws UndefiniertException {
		final List<StrassenSegment> liste = new ArrayList<>();
		liste.addAll(getSituationsDaten().getSegmente());
		return liste;
	}

	/**
	 * liefert den aktuellen Datensatz mit den Situationsparametern. Wenn keine
	 * gültigen Parameter verfügbar sind wird eine {@link UndefiniertException}
	 * geworfen.
	 *
	 * @return die Daten
	 * @throws UndefiniertException
	 *             es konnten keine gültigen Parameter ermittelt werden
	 */
	private PdSituationsEigenschaften.Daten getSituationsDaten() throws UndefiniertException {
		final PdSituationsEigenschaften.Daten daten = baustelle.getParameterDatensatz(PdSituationsEigenschaften.class)
				.getDatum();
		if ((daten == null) || (!daten.isValid())) {
			throw new UndefiniertException("Die Daten der Baustelle " + baustelle + " sind nicht bekannt");
		}
		return daten;
	}

	@Override
	public final long getStart() throws UndefiniertException {
		return getSituationsDaten().getStartZeit();
	}

	/**
	 * Die Funktion ermittelt den aktuellen Status der Baustelle aus dem
	 * aktuellen Situationsparameterdatensatz. Wenn kein passender Status
	 * gefunden wurde bzw. der Parameterdatensatz nicht verfügbar ist, wird der
	 * Status ENTWORFEN geliefert.
	 *
	 * @return den aktuellen Status oder ENTWORFEN, wenn keiner ermittelt werden
	 *         konnte.
	 */
	public final BaustellenStatus getStatus() {
		BaustellenStatus status = BaustellenStatus.ENTWORFEN;
		try {
			status = getBaustellenDaten().getBaustellenStatus();
		} catch (final UndefiniertException e) {
			Debug.getLogger().fine(e.getLocalizedMessage());
			// es wird der Standardwert ENTWORFEN geliefert
		}
		return status;
	}

}
