/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.verwaltung;

/**
 * die Repräsentation eines Stauobjekts innerhalb der Stauprognose eine
 * Baustelle.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class BssStau {

	/**
	 * der Zeitpunkt des Staubeginns.
	 */
	private final long startZeit;

	/**
	 * der Zeitpunkt des Stauendes.
	 */
	private long endZeit;

	/**
	 * die maximale Länge das Staus im Verlauf seiner Existenz.
	 */
	private long maximaleLaenge;

	/**
	 * der Zeitpunkt zu dem der Stau seine maximale Länge erreicht hat.
	 */
	private long maximaleLaengeZeit;

	/**
	 * die maximale Verlustzeit innerhalb des Staus.
	 */
	private long verlustZeit;

	/**
	 * erzeugt eine Instanz des Stauobjekts.
	 *
	 * @param zeit
	 *            der Zeitpunkt der Entstehung des Staus
	 * @param laenge
	 *            die Länge des Staus bei der Entstehung
	 * @param verlustZeit
	 *            die Verlustzeit zum Entstehungszeitpunkt
	 */
	public BssStau(final long zeit, final long laenge, final long verlustZeit) {
		this.startZeit = zeit;
		this.endZeit = zeit;
		aktualisieren(zeit, laenge, verlustZeit);
	}

	/**
	 * setzt die Verlustzeit des Staus.<br>
	 * Das Stauobjekt speichert die maximale Verlustzeit im Verlauf der Existenz
	 * des Staus. Der übergebene Wert wird deshalb nur übernommen, wenn der
	 * bereits gespeicherte kleiner ist.
	 *
	 * @param zeit
	 *            die neu zu setzende Verlustzeit
	 */
	private void setVerlustZeit(final long zeit) {
		if (isBeendet()) {
			throw new IllegalStateException(
					"Es wurde versucht die Verlustzeit an einem beendeten Stauobjekt zu setzen");
		}
		verlustZeit = Math.max(verlustZeit, zeit);
	}

	/**
	 * ermittelt, ob der Stau beendet wurde.<br>
	 * Ein Stau gilt als beendet, wenn er eine Dauer &gt; 0 hat, d.h. wenn de
	 * Endzeit größer als die Anfangszeit ist.
	 *
	 * @return true, wenn der Stau beendet ist
	 */
	private boolean isBeendet() {
		return endZeit > startZeit;
	}

	/**
	 * setzt die Länge des Staus und den dazugehörigen Zeitpunkt.<br>
	 * Das Stauobjekt speichert die maximale Länge im Verlauf der Existenz des
	 * Staus. Der übergebene Wert wird deshalb nur übernommen, wenn der bereits
	 * gespeicherte kleiner ist.
	 *
	 * @param zeit
	 *            der Zeitpunkt zu dem die übergeben Steulänge erreicht wurde
	 * @param laenge
	 *            die Länge
	 */
	private void setMaximaleLaenge(final long zeit, final long laenge) {
		if (isBeendet()) {
			throw new IllegalStateException("Es wurde versucht die Staulänge an einem beendeten Stauobjekt zu setzen");
		}
		if (maximaleLaenge < Math.max(0, laenge)) {
			maximaleLaengeZeit = zeit;
			maximaleLaenge = laenge;
		}
	}

	/**
	 * liefert den Anfangszeitpunkt des Staus.
	 *
	 * @return den Zeitpunkt
	 */
	public final long getStartZeit() {
		return startZeit;
	}

	/**
	 * liefert die Dauer des Staus in Millisekunden.<br>
	 * Wenn kein Endzeitpunkt bestimmt wurde, wird der Wert 0 geliefert.
	 *
	 * @return den Zeitpunkt oder 0
	 */
	public final long getDauer() {
		return endZeit - startZeit;
	}

	/**
	 * liefert die maximale Länge, die während der Existenz des Staus
	 * aufgetreten ist.
	 *
	 * @return die Länge
	 */
	public final long getMaxLaenge() {
		return maximaleLaenge;
	}

	/**
	 * liefert den Zeitpunkt zu dem der Stau seine maximale Länge erreicht hat.
	 *
	 * @return den Zeitpunkt
	 */
	public final long getMaxLaengeZeit() {
		return maximaleLaengeZeit;
	}

	/**
	 * liefert die maximale Verlustzeit im Verlauf des Staus.
	 *
	 * @return die Zeitdauer
	 */
	public final long getVerlustZeit() {
		return verlustZeit;
	}

	/**
	 * markiert den Stau als beendet und setzt die Endezeit.
	 *
	 * @param zs
	 *            der Zeitpunkt für das Stauende
	 */
	public final void beenden(final long zs) {
		endZeit = zs;
	}

	/**
	 * aktualisiert die Daten des Stauobjekts.<br>
	 * Die Wert für Länge und Verlustzeit werden entsprechend in das Stauobjekt
	 * eingebracht.
	 *
	 * @param zs
	 *            der zugehörige Zeitpunkt
	 * @param laenge
	 *            die neue Staulänge
	 * @param zeit
	 *            die neue Verlustzeit
	 */
	public final void aktualisieren(final long zs, final long laenge, final long zeit) {
		setMaximaleLaenge(zs, laenge);
		setVerlustZeit(zeit);
	}

}
